from musiclang_predict import MusicLangTokenizer
from musiclang.library import *


default_options = {
    'chord_change_token': True,
    'melody_end_token': True,
    'chord_duration_token': True,
    'density_token': True,
    'chord_extension_token': True,
    'next_chord_token': True,
    'next_chord_duration_token': True,
    'will_end_token': True,
    'dissonance_token': True,
    'amplitude_token': True,
    'average_octave_token': True,
    'voice_token': True,
    'random_instrument_permutation': False
}

def test_tokenizer():

    pass



def test_cut_bar():
    pass

def test_add_bar():
    pass